@@ -1,41 +1,33 @@
 ##
-# copyright 2009, James William Pye
-# http://python.projects.postgresql.org
+# .api - ABCs for database interface elements
 ##
 """
-Application Programmer Interface specifications for PostgreSQL (ABCs).
+Application Programmer Interfaces for PostgreSQL.
 
-PG-API
-======
+``postgresql.api`` is a collection of Python APIs for the PostgreSQL DBMS. It
+is designed to take full advantage of PostgreSQL's features to provide the
+Python programmer with substantial convenience.
 
-``postgresql.api`` is a Python API for the PostgreSQL DBMS. It is designed to take
-full advantage of PostgreSQL's features to provide the Python programmer with
-substantial convenience.
-
-This module is used to define PG-API. It creates a set of ABCs
+This module is used to define "PG-API". It creates a set of ABCs
 that makes up the basic interfaces used to work with a PostgreSQL server.
 """
-import os
-import sys
-import warnings
 import collections
 from abc import abstractproperty, abstractmethod
-from operator import itemgetter
 
-from . import sys as pg_sys
-from .python.element import Element, prime_factor
+from .python.element import Element
 from .python.doc import Doc
 from .python.decorlib import propertydoc
 
 __all__ = [
 	'Message',
-	'PreparedStatement',
+	'Statement',
 	'Chunks',
 	'Rows',
 	'Cursor',
 	'Connector',
 	'Category',
 	'Database',
+	'TypeIO',
 	'Connection',
 	'Transaction',
 	'Settings',
@@ -47,12 +39,10 @@ __all__ = [
 
 class Message(Element):
 	"""
-	A message emitted by PostgreSQL. This element is universal, so
-	`postgresql.api.Message` is a complete implementation for representing a
-	message. Any interface should produce these objects.
+	A message emitted by PostgreSQL.
+	A message being a NOTICE, WARNING, INFO, etc.
 	"""
-	_e_label = property(lambda x: getattr(x, 'details').get('severity', 'MESSAGE'))
-	_e_factors = ('creator',)
+	_e_label = 'MESSAGE'
 
 	severities = (
 		'DEBUG',
@@ -68,101 +58,56 @@ class Message(Element):
 		'CLIENT',
 	)
 
-	# What generated the message?
-	source = 'SERVER'
-
-	code = "00000"
-	message = None
-	details = None
-
-	def __init__(self,
-		message : "The primary information of the message",
-		code : "Message code to attach (SQL state)" = None,
-		details : "additional information associated with the message" = {},
-		source : "Which side generated the message(SERVER, CLIENT)" = None,
-		creator : "The interface element that called for instantiation" = None,
-	):
-		self.message = message
-		self.details = details
-		self.creator = creator
-		if code is not None and self.code != code:
-			self.code = code
-		if source is not None and self.source != source:
-			self.source = source
-
-	def __repr__(self):
-		return "{mod}.{typname}({message!r}{code}{details}{source}{creator})".format(
-			mod = self.__module__,
-			typname = self.__class__.__name__,
-			message = self.message,
-			code = (
-				"" if self.code == type(self).code
-				else ", code = " + repr(self.code)
-			),
-			details = (
-				"" if not self.details
-				else ", details = " + repr(self.details)
-			),
-			source = (
-				"" if self.source is None
-				else ", source = " + repr(self.source)
-			),
-			creator = (
-				"" if self.creator is None
-				else ", creator = " + repr(self.creator)
-			)
-		)
-
-	@property
-	def location_string(self):
-		details = self.details
-		loc = [
-			details.get(k, '?') for k in ('file', 'line', 'function')
-		]
-		return (
-			"" if loc == ['?', '?', '?']
-			else "File {0!r}, "\
-			"line {1!s}, in {2!s}".format(*loc)
-		)
-
-	# keys to filter in .details
-	standard_detail_coverage = frozenset(['message', 'severity', 'file', 'function', 'line',])
-
-	def _e_metas(self):
-		yield (None, self.message)
-		if self.code and self.code != "00000":
-			yield ('CODE', self.code)
-		locstr = self.location_string
-		if locstr:
-			yield ('LOCATION', locstr + ' from ' + self.source)
-		else:
-			yield ('LOCATION', self.source)
-		for k, v in sorted(self.details.items(), key = itemgetter(0)):
-			if k not in self.standard_detail_coverage:
-				yield (k.upper(), str(v))
-
-	def raise_message(self, starting_point = None):
+	@abstractproperty
+	def source(self) -> str:
 		"""
-		Take the given message object and hand it to all the primary
-		factors(creator) with a msghook callable.
+		Where the message originated from. Normally, 'SERVER', but sometimes
+		'CLIENT'.
 		"""
-		if starting_point is not None:
-			f = starting_point
-		else:
-			f = self.creator
 
-		while f is not None:
-			if getattr(f, 'msghook', None) is not None:
-				if f.msghook(self):
-					# the trap returned a nonzero value,
-					# so don't continue raising. (like with's __exit__)
-					return f
-			f = prime_factor(f)
-			if f:
-				f = f[1]
-		# if the next primary factor is without a raise or does not exist,
-		# send the message to postgresql.sys.msghook
-		pg_sys.msghook(self)
+	@abstractproperty
+	def code(self) -> str:
+		"""
+		The SQL state code of the message.
+		"""
+
+	@abstractproperty
+	def message(self) -> str:
+		"""
+		The primary message string.
+		"""
+
+	@abstractproperty
+	def details(self) -> dict:
+		"""
+		The additional details given with the message. Common keys *should* be the
+		following:
+
+		 * 'severity'
+		 * 'context'
+		 * 'detail'
+		 * 'hint'
+		 * 'file'
+		 * 'line'
+		 * 'function'
+		 * 'position'
+		 * 'internal_position'
+		 * 'internal_query'
+		"""
+
+	@abstractmethod
+	def isconsistent(self, other) -> bool:
+		"""
+		Whether the fields of the `other` Message object is consistent with the
+		fields of `self`.
+
+		This *must* return the result of the comparison of code, source, message,
+		and details.
+
+		This method is provided as the alternative to overriding equality;
+		often, pointer equality is the desirable means for comparison, but
+		equality of the fields is also necessary.
+		"""
 
 class Result(Element):
 	"""
@@ -248,7 +193,7 @@ class Result(Element):
 
 	@propertydoc
 	@abstractproperty
-	def statement(self) -> ("PreparedStatement", None):
+	def statement(self) -> ("Statement", None):
 		"""
 		The query object used to create the cursor. `None`, if unknown.
 
@@ -273,7 +218,7 @@ class Cursor(
 	set. Cursors publish a file-like interface for reading tuples from a cursor
 	declared on the database.
 
-	`Cursor` objects are created by invoking the `PreparedStatement.declare`
+	`Cursor` objects are created by invoking the `Statement.declare`
 	method or by opening a cursor using an identifier via the
 	`Database.cursor_from_id` method.
 	"""
@@ -354,18 +299,19 @@ class Cursor(
 		FROM_END will effectively be ABSOLUTE.
 		"""
 
-class PreparedStatement(
+class Statement(
 	Element,
 	collections.Callable,
 	collections.Iterable,
 ):
 	"""
-	Instances of `PreparedStatement` are returned by the `prepare` method of
+	Instances of `Statement` are returned by the `prepare` method of
 	`Database` instances.
 
-	A PreparedStatement is an Iterable as well as Callable. This feature is
-	supported for queries that have the default arguments filled in or take no
-	arguments at all. It allows for things like:
+	A Statement is an Iterable as well as Callable.
+
+	The Iterable interface is supported for queries that take no arguments at
+	all. It allows the syntax::
 
 		>>> for x in db.prepare('select * FROM table'):
 		...  pass
@@ -470,7 +416,7 @@ class PreparedStatement(
 		"""
 
 	@abstractmethod
-	def clone(self) -> "PreparedStatement":
+	def clone(self) -> "Statement":
 		"""
 		Create a new statement object using the same factors as `self`.
 
@@ -490,23 +436,6 @@ class PreparedStatement(
 		[...]
 		"""
 
-	@abstractmethod
-	def rows(self, *parameters) -> collections.Iterable:
-		"""
-		Return an iterator producing rows produced by the cursor
-		created from the statement bound with the given parameters.
-
-		Row iterators are never scrollable.
-
-		Supporting cursors will be WITH HOLD when outside of a transaction.
-
-		`rows` is designed for the situations involving large data sets.
-
-		Each iteration returns a single row. Arguably, best implemented:
-
-			return itertools.chain.from_iterable(self.chunks(*parameters))
-		"""
-
 	@abstractmethod
 	def chunks(self, *parameters) -> collections.Iterable:
 		"""
@@ -517,11 +446,49 @@ class PreparedStatement(
 
 		Supporting cursors will be WITH HOLD when outside of a transaction.
 
-		`chunks` is designed for the situations involving large data sets.
+		`chunks` is designed for moving large data sets efficiently.
 
 		Each iteration returns sequences of rows *normally* of length(seq) ==
 		chunksize. If chunksize is unspecified, a default, positive integer will
-		be filled in.
+		be filled in. The rows contained in the sequences are only required to
+		support the basic `collections.Sequence` interfaces; simple and quick
+		sequence types should be used.
+		"""
+
+	@abstractmethod
+	def rows(self, *parameters) -> collections.Iterable:
+		"""
+		Return an iterator producing rows produced by the cursor
+		created from the statement bound with the given parameters.
+
+		Row iterators are never scrollable.
+
+		Supporting cursors will be WITH HOLD when outside of a transaction to
+		allow cross-transaction access.
+
+		`rows` is designed for the situations involving large data sets.
+
+		Each iteration returns a single row. Arguably, best implemented::
+
+			return itertools.chain.from_iterable(self.chunks(*parameters))
+		"""
+
+	@abstractmethod
+	def column(self, *parameters) -> collections.Iterable:
+		"""
+		Return an iterator producing the values of the first column in
+		the cursor created from the statement bound with the given parameters.
+
+		Column iterators are never scrollable.
+
+		Supporting cursors will be WITH HOLD when outside of a transaction to
+		allow cross-transaction access.
+
+		`column` is designed for the situations involving large data sets.
+
+		Each iteration returns a single value. `column` is equivalent to::
+
+			return map(operator.itemgetter(0), self.rows(*parameters))
 		"""
 
 	@abstractmethod
@@ -595,6 +562,7 @@ class PreparedStatement(
 		"""
 		Close the prepared statement releasing resources associated with it.
 		"""
+PreparedStatement = Statement
 
 class StoredProcedure(
 	Element,
@@ -651,21 +619,6 @@ class Transaction(Element):
 		>>> with db.xact():
 		...  with db.xact():
 		...   ...
-
-	[DEPRECATED]
-
-	Or, in cases where two-phase commit is desired:
-
-		>>> with db.xact(gid = 'gid') as gxact:
-		...  with gxact:
-		...   # phase 1 block
-		...   ...
-		>>> # fully committed at this point
-
-	Considering that transactions decide what's saved and what's not saved, it is
-	important that they are used properly. In most situations, when an action is
-	performed where state of the transaction is unexpected, an exception should
-	occur.
 	"""
 	_e_label = 'XACT'
 	_e_factors = ('database',)
@@ -694,20 +647,6 @@ class Transaction(Element):
 		the START TRANSACTION statement.
 		"""
 
-	@propertydoc
-	@abstractproperty
-	def gid(self) -> (None, str):
-		"""
-		[DEPRECATED]
-
-		The global identifier of the transaction block:
-
-			PREPARE TRANSACTION <gid>;
-
-		The `gid` property is a string that indicates that the block is a prepared
-		transaction.
-		"""
-
 	@abstractmethod
 	def start(self) -> None:
 		"""
@@ -726,7 +665,7 @@ class Transaction(Element):
 		`self.mode` specifies the mode of the transaction. Normally, ``READ
 		ONLY`` or ``READ WRITE``.
 
-		If the transaction is open--started or prepared, do nothing.
+		If the transaction is already open, do nothing.
 
 		If the transaction has been committed or aborted, raise an
 		`postgresql.exceptions.OperationError`.
@@ -738,9 +677,6 @@ class Transaction(Element):
 		"""
 		Commit the transaction.
 
-		If the transaction is configured with a `gid` issue a COMMIT PREPARED
-		statement with the configured `gid`.
-
 		If the transaction is a block, issue a COMMIT statement.
 
 		If the transaction was started inside a transaction block, it should be
@@ -759,46 +695,9 @@ class Transaction(Element):
 		If the transaction is a transaction block, issue an ABORT.
 
 		If the transaction has already been aborted, do nothing.
-
-		[DEPRECATED]
-		If the transaction is configured with a `gid` *and* has been prepared, issue
-		a ROLLBACK PREPARE statement with the configured `gid`.
 		"""
 	abort = rollback
 
-	@abstractmethod
-	def recover(self) -> None:
-		"""
-		[DEPRECATED]
-
-		If the transaction is assigned a `gid`, recover may be used to identify
-		the transaction as prepared and ready for committing or aborting.
-
-		This method is used in recovery procedures where a prepared transaction
-		needs to be committed or rolled back.
-
-		If no prepared transaction with the configured `gid` exists, a
-		`postgresql.exceptions.UndefinedObjectError` must be raised.
-		[This is consistent with the error raised by ROLLBACK/COMMIT PREPARED]
-
-		Once this method has been ran, it should identify the transaction as being
-		prepared so that subsequent invocations to `commit` or `rollback` should
-		cause the appropriate ROLLBACK PREPARED or COMMIT PREPARED statements to
-		be executed.
-		"""
-
-	@abstractmethod
-	def prepare(self) -> None:
-		"""
-		[DEPRECATED]
-
-		Explicitly prepare the transaction with the configured `gid` by issuing a
-		PREPARE TRANSACTION statement with the configured `gid`.
-		This *must* be called for the first phase of the commit.
-
-		If the transaction is already prepared, do nothing.
-		"""
-
 	@abstractmethod
 	def __enter__(self):
 		"""
@@ -822,9 +721,7 @@ class Transaction(Element):
 		unavailable, the `rollback` method should cause a
 		`postgresql.exceptions.ConnectionDoesNotExistError` exception to occur.
 
-		Otherwise, run the transaction's `commit` method. If the commit fails,
-		a `gid` is configured, and the connection is still available, run the
-		transaction's `rollback` method.
+		Otherwise, run the transaction's `commit` method.
 
 		When the `commit` is ultimately unsuccessful or not ran at all, the purpose
 		of __exit__ is to resolve the error state of the database iff the
@@ -955,7 +852,6 @@ class Database(Element):
 	@propertydoc
 	@abstractproperty
 	def xact(self,
-		gid : "global identifier to configure" = None,
 		isolation : "ISOLATION LEVEL to use with the transaction" = None,
 		mode : "Mode of the transaction, READ ONLY or READ WRITE" = None,
 	) -> Transaction:
@@ -972,16 +868,25 @@ class Database(Element):
 		"""
 
 	@abstractmethod
-	def execute(sql) -> None:
+	def do(language, source) -> None:
 		"""
-		Execute an arbitrary block of SQL. Always returns `None` and raise
-		a `postgresql.exceptions.Error` subclass on error.
+		Execute a DO statement using the given language and source.
+		Always returns `None`.
+
+		Likely to be a function of Connection.execute.
 		"""
 
 	@abstractmethod
-	def prepare(self, sql : str) -> PreparedStatement:
+	def execute(sql) -> None:
 		"""
-		Create a new `PreparedStatement` instance bound to the connection
+		Execute an arbitrary block of SQL. Always returns `None` and raise
+		an exception on error.
+		"""
+
+	@abstractmethod
+	def prepare(self, sql : str) -> Statement:
+		"""
+		Create a new `Statement` instance bound to the connection
 		using the given SQL.
 
 		>>> s = db.prepare("SELECT 1")
@@ -993,9 +898,9 @@ class Database(Element):
 	@abstractmethod
 	def statement_from_id(self,
 		statement_id : "The statement's identification string.",
-	) -> PreparedStatement:
+	) -> Statement:
 		"""
-		Create a `PreparedStatement` object that was already prepared on the
+		Create a `Statement` object that was already prepared on the
 		server. The distinction between this and a regular query is that it
 		must be explicitly closed if it is no longer desired, and it is
 		instantiated using the statement identifier as opposed to the SQL
@@ -1013,7 +918,7 @@ class Database(Element):
 		`Cursor` objects created this way must *not* be closed when the object
 		is garbage collected. Rather, the user must explicitly close it for
 		the server resources to be released. This is in contrast to `Cursor`
-		objects that are created by invoking a `PreparedStatement` or a SRF
+		objects that are created by invoking a `Statement` or a SRF
 		`StoredProcedure`.
 		"""
 
@@ -1054,6 +959,68 @@ class Database(Element):
 		is being recycled.
 		"""
 
+	@abstractmethod
+	def notify(self, *channels, **channel_and_payload) -> int:
+		"""
+		NOTIFY the channels with the given payload.
+
+		Equivalent to issuing "NOTIFY <channel>" or "NOTIFY <channel>, <payload>"
+		for each item in `channels` and `channel_and_payload`. All NOTIFYs issued
+		*must* occur in the same transaction.
+
+		The items in `channels` can either be a string or a tuple. If a string,
+		no payload is given, but if an item is a `builtins.tuple`, the second item
+		will be given as the payload. `channels` offers a means to issue NOTIFYs
+		in guaranteed order.
+
+		The items in `channel_and_payload` are all payloaded NOTIFYs where the
+		keys are the channels and the values are the payloads. Order is undefined.
+		"""
+
+	@abstractmethod
+	def listen(self, *channels) -> None:
+		"""
+		Start listening to the given channels.
+
+		Equivalent to issuing "LISTEN <x>" for x in channels.
+		"""
+
+	@abstractmethod
+	def unlisten(self, *channels) -> None:
+		"""
+		Stop listening to the given channels.
+
+		Equivalent to issuing "UNLISTEN <x>" for x in channels.
+		"""
+
+	@abstractmethod
+	def listening_channels(self) -> ["channel name", ...]:
+		"""
+		Return an *iterator* to all the channels currently being listened to.
+		"""
+
+	@abstractmethod
+	def iternotifies(self, timeout = None) -> collections.Iterator:
+		"""
+		Return an iterator to the notifications received by the connection. The
+		iterator *must* produce triples in the form ``(channel, payload, pid)``.
+
+		If timeout is not `None`, `None` *must* be emitted at the specified
+		timeout interval. If the timeout is zero, all the pending notifications
+		*must* be yielded by the iterator and then `StopIteration` *must* be
+		raised.
+
+		If the connection is closed for any reason, the iterator *must* silently
+		stop by raising `StopIteration`. Further error control is then the
+		responsibility of the user.
+		"""
+
+class TypeIO(Element):
+	_e_label = 'TYPIO'
+
+	def _e_metas(self):
+		return ()
+
 class SocketFactory(object):
 	@propertydoc
 	@abstractproperty
@@ -1258,7 +1225,7 @@ class Installation(Element):
 	information about an installation of PostgreSQL accessible by the Python 
 	"""
 	_e_label = "INSTALLATION"
-	_e_factors = ('pg_config_path',)
+	_e_factors = ()
 
 	@propertydoc
 	@abstractproperty
