@@ -8,20 +8,58 @@ import os
 import pprint
 from struct import unpack, Struct
 from .message_types import message_types
-from .typstruct import ushort_pack, ushort_unpack, ulong_pack, ulong_unpack
-
-def pack_tuple_data(atts):
-	return b''.join([
-		b'\xff\xff\xff\xff'
-		if x is None
-		else (ulong_pack(len(x)) + x)
-		for x in atts
-	])
+from ..python.structlib import ushort_pack, ushort_unpack, ulong_pack, ulong_unpack
 
 try:
-	from .optimized import parse_tuple_message, pack_tuple_data
+	from ..port.optimized import parse_tuple_message, pack_tuple_data
 except ImportError:
-	pass
+	def pack_tuple_data(atts,
+		none = None,
+		ulong_pack = ulong_pack,
+		blen = bytes.__len__
+	):
+		return b''.join([
+			b'\xff\xff\xff\xff'
+			if x is none
+			else (ulong_pack(blen(x)) + x)
+			for x in atts
+		])
+
+try:
+	from ..port.optimized import cat_messages
+except ImportError:
+	from ..python.structlib import lH_pack, long_pack
+	# Special case tuple()'s
+	def _pack_tuple(t,
+		blen = bytes.__len__,
+		tlen = tuple.__len__,
+		pack_head = lH_pack,
+		ulong_pack = ulong_pack,
+		ptd = pack_tuple_data,
+	):
+		# NOTE: duplicated from above
+		r = b''.join([
+			b'\xff\xff\xff\xff'
+			if x is None
+			else (ulong_pack(blen(x)) + x)
+			for x in t
+		])
+		return pack_head((blen(r) + 6, tlen(t))) + r
+
+	def cat_messages(messages,
+		lpack = long_pack,
+		blen = bytes.__len__,
+		tuple = tuple,
+		pack_tuple = _pack_tuple
+	):
+		return b''.join([
+			(x.bytes() if x.__class__ is not bytes else (
+				b'd' + lpack(blen(x) + 4) + x
+			)) if x.__class__ is not tuple else (
+				b'D' + pack_tuple(x)
+			) for x in messages
+		])
+	del _pack_tuple, lH_pack, long_pack
 
 StringFormat = b'\x00\x00'
 BinaryFormat = b'\x00\x01'
@@ -158,23 +196,23 @@ class EmptyMessage(Message):
 class Notify(Message):
 	'Asynchronous notification message'
 	type = message_types[b'A'[0]]
-	__slots__ = ('pid', 'relation', 'parameter')
+	__slots__ = ('pid', 'channel', 'payload',)
 
-	def __init__(self, pid, relation, parameter = b''):
+	def __init__(self, pid, channel, payload = b''):
 		self.pid = pid
-		self.relation = relation
-		self.parameter = parameter
+		self.channel = channel
+		self.payload = payload
 
 	def serialize(self):
 		return ulong_pack(self.pid) + \
-			self.relation + b'\x00' + \
-			self.parameter + b'\x00'
+			self.channel + b'\x00' + \
+			self.payload + b'\x00'
 
 	@classmethod
 	def parse(typ, data):
-		pid = ulong_unpack(data[0:4])
-		relname, param, nothing = data[4:].split(b'\x00', 2)
-		return typ(pid, relname, param)
+		pid = ulong_unpack(data)
+		channel, payload, _ = data[4:].split(b'\x00', 2)
+		return typ(pid, channel, payload)
 
 class ShowOption(Message):
 	"""ShowOption(name, value)
@@ -206,18 +244,19 @@ class Complete(StringMessage):
 		"""
 		Extract the last set of digits as an integer.
 		"""
-		rms = self.data.strip().split()
-		if rms[0].lower() == b'copy':
-			if len(rms) > 1:
-				return int(rms[-1])
-		elif rms[-1].isdigit():
-			return int(rms[-1])
+		# Find the last sequence of digits.
+		# If there are no fields consisting only of digits, there is no count.
+		for x in reversed(self.data.split()):
+			if x.isdigit():
+				return int(x)
+		return None
 
 	def extract_command(self):
-		t = self.data.strip().split()
-		if t:
-			return t[0]
-		return None
+		"""
+		Strip all the *surrounding* digits and spaces from the command tag,
+		and return that string.
+		"""
+		return self.data.strip(b'\c\n\t 0123456789') or None
 
 class Null(EmptyMessage):
 	'Null command'
@@ -280,60 +319,29 @@ class Ready(Message):
 		return self.xact_state
 
 class Notice(Message, dict):
-	"""Notification message"""
-	type = message_types[b'N'[0]]
-	_dtm = {
-		b'S' : 'severity',
-		b'C' : 'code',
-		b'M' : 'message',
-		b'D' : 'detail',
-		b'H' : 'hint',
-		b'W' : 'context',
-		b'P' : 'position',
-		b'p' : 'internal_position',
-		b'q' : 'internal_query',
-		b'F' : 'file',
-		b'L' : 'line',
-		b'R' : 'function',
-	}
-	__slots__ = ()
+	"""
+	Notification message
 
-	def __init__(self,
-		severity = None,
-		message = None,
-		code = None,
-		detail = None,
-		hint = None,
-		position = None,
-		internal_position = None,
-		internal_query = None,
-		context = None,
-		file = None,
-		line = None,
-		function = None
-	):
-		for (k, v) in locals().items():
-			if v not in (self, None):
-				self[k] = v
+	Used by PQ to emit INFO, NOTICE, and WARNING messages among other
+	severities.
+	"""
+	type = message_types[b'N'[0]]
+	__slots__ = ()
 	__repr__ = dict_message_repr
 
 	def serialize(self):
-		return b''.join([
-			k + self[v] + b'\x00'
-			for k, v in self._dtm.items()
-			if self.get(v) is not None
+		return b'\x00'.join([
+			k + v for k, v in self.items()
+			if k and v is not None
 		]) + b'\x00'
 
 	@classmethod
-	def parse(typ, data):
-		kw = {}
-		g = typ._dtm.get
-		for frag in data.split(b'\x00'):
-			if frag:
-				key = g(frag[0:1])
-				if key is not None:
-					kw[key] = frag[1:]
-		return typ(**kw)
+	def parse(typ, data, msgtypes = message_types):
+		return typ([
+			(msgtypes[x[0]], x[1:])
+			# "if x" reduce empty fields
+			for x in data.split(b'\x00') if x
+		])
 
 class ClientNotice(Notice):
 	__slots__ = ()
@@ -371,7 +379,7 @@ class FunctionResult(Message):
 	def serialize(self):
 		return self.result is None and b'\xff\xff\xff\xff' or \
 			ulong_pack(len(self.result)) + self.result
-	
+
 	@classmethod
 	def parse(typ, data):
 		if data == b'\xff\xff\xff\xff':
@@ -443,10 +451,14 @@ class Tuple(TupleMessage):
 		return ushort_pack(len(self)) + pack_tuple_data(self)
 
 	@classmethod
-	def parse(typ, data):
+	def parse(typ, data,
+		T = tuple, ulong_unpack = ulong_unpack,
+		len = len
+	):
 		natts = ushort_unpack(data[0:2])
-		atts = list()
+		atts = []
 		offset = 2
+		add = atts.append
 
 		while natts > 0:
 			alo = offset
@@ -459,12 +471,13 @@ class Tuple(TupleMessage):
 				ao = offset
 				offset = ao + al
 				att = data[ao:offset]
-			atts.append(att)
+			add(att)
 			natts -= 1
-		return typ(atts)
+		return T(atts)
 	try:
-		parse = classmethod(parse_tuple_message)
+		parse = parse_tuple_message
 	except NameError:
+		# This is an override when port.optimized is available.
 		pass
 
 class KillInformation(Message):
@@ -687,15 +700,14 @@ class Bind(Message):
 		self.arguments = arguments
 		self.rformats = rformats
 
-	def serialize(self):
+	def serialize(self, len = len):
 		args = self.arguments
 		ac = ushort_pack(len(args))
 		ad = pack_tuple_data(tuple(args))
-		rfc = ushort_pack(len(self.rformats))
 		return \
 			self.name + b'\x00' + self.statement + b'\x00' + \
-			ac + b''.join(self.aformats) + ac + ad + rfc + \
-			b''.join(self.rformats)
+			ac + b''.join(self.aformats) + ac + ad + \
+			ushort_pack(len(self.rformats)) + b''.join(self.rformats)
 
 	@classmethod
 	def parse(typ, message_data):
@@ -805,7 +817,7 @@ class Function(Message):
 	type = message_types[b'F'[0]]
 	__slots__ = ('oid', 'aformats', 'arguments', 'rformat')
 
-	def __init__(self, oid, aformats, args, rformat = StringFormat):
+	def __init__(self, oid, aformats, args, rformat):
 		self.oid = oid
 		self.aformats = aformats
 		self.arguments = args
